////////////////////////////////////////////////////////////////////////////////
// File         : RelationBaustein.cpp
// Subsystem    : Graph
// Component    : object networks
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class RelationBaustein
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 10.10.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////


//#define TRACE
#define TRACE_BENES

#include "RelationBaustein.h"
#include "SnetPrint.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

RelationBaustein::RelationBaustein( bool injective, IsoTypeNode &class_isoType, int numberOfEntries, int numberOfExits, Baustein *owner, HnetUniverse *universe )
	: m_injectiveRelation( injective ), m_exitEntry( numberOfExits, NO_ENTRY ), Baustein( class_isoType, numberOfEntries, numberOfExits, owner, universe )
{

}

RelationBaustein::~RelationBaustein()
{

}

//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

void RelationBaustein::PrepareRelationComputation( void )
{
	// for optimisation later use m_snetComplement instead 
//	IsoTypeNode iso = GetIsoType();
//	IsoTypeNode isoCopy = m_universe->CopyIsoType( iso );

//	Hnet &hnet = m_universe->GetHnet( isoCopy );
//	hnet.MakeFlat();
//	Snet m_snet = hnet.GetSnet(); 

//	PrepareOuterEntriesExits( m_snetComplement );
//	Snet &snet = hnet.GetSnet();
//	snet.TransComplement( m_snetComplement, m_originalNode );
//	PrepareOuterEntriesExits( m_snetComplement );
}

bool RelationBaustein::DoRelationComputation( int defaultColor, int numberOfEntryColors, int *entryColors, int numberOfExitColors, int *exitColors )
{
	int i, min;
	bool ret;
	NodeAttribute< int > color( defaultColor );
	NodeAttribute< bool > multible( false );
	NodeAttribute< Relation > relation;

	IsoTypeNode iso = GetIsoType();
	IsoTypeNode isoCopy = m_universe->CopyIsoType( iso );

	Hnet &hnet = m_universe->GetHnet( isoCopy );

	MapSubcentresElementary( hnet, hnet.FirstSubcentre() ); // map subcentre to every elementary subbaustein	

	Snet &snet = hnet.GetSnet();

#ifdef TRACE
	NodeAttribute< int > colorP( defaultColor );
	NodeAttribute< bool > multibleP( false );
	NodeAttribute< Relation > relationP;

	SnetPrint snetP( snet, "flat snet" );
	snetP.Print( );

	PrepareOuterEntriesExits( snetP );

	min = (numberOfEntryColors <= NumberOfEntries()) ? numberOfEntryColors : NumberOfEntries();
	for ( i = 0; i < min ; i++ )
	{
		colorP[ OuterEntry( i ) ] = entryColors[ i ];
	}

	min = (numberOfExitColors <= NumberOfExits()) ? numberOfExitColors : NumberOfExits();
	for ( i = 0; i < min; i++ )
	{
		colorP[ OuterExit( i ) ] = exitColors[ i ];
	}

	SetMultibleFlags( multibleP, snetP );
	ret = snetP.ComputeRelations( colorP, multibleP, relationP );

	snetP.PrintRelation( relationP );
#endif  //#ifdef TRACE

	PrepareOuterEntriesExits( snet );

	min = (numberOfEntryColors <= NumberOfEntries()) ? numberOfEntryColors : NumberOfEntries();
	for ( i = 0; i < min ; i++ )
	{
		color[ OuterEntry( i ) ] = entryColors[ i ];
	}

	min = (numberOfExitColors <= NumberOfExits()) ? numberOfExitColors : NumberOfExits();
	for ( i = 0; i < min; i++ )
	{
		color[ OuterExit( i ) ] = exitColors[ i ];
	}

	SetMultibleFlags( multible, snet );
	ret = snet.ComputeRelations( color, multible, relation );

	// for optimisation later use m_snetComplement instead 
	//	ret = m_snetComplement.ComputeRelations( color, relation ); 

	if (ret )
		SetRelationsElementary( color, relation );

	return ret;
}


void RelationBaustein::FinishRelationComputation( void )
{
	// for optimisation later use m_snetComplement instead 
//	m_snetComplement = Snet();
}


void RelationBaustein::SetRelationsElementary( NodeAttribute< int > &color, const NodeAttribute< Relation > &relation )
{
	if ( m_elementary )
	{
		SetRelation( color, relation[ m_subcentre ] );
		return; 
	}
	else
	{
		for( int i = 0; i <	m_numberOfSubbausteins; i++)
			 reinterpret_cast< RelationBaustein *>( m_subbausteins[ i ] )->SetRelationsElementary( color, relation );
	}
}


void RelationBaustein::SetMultibleFlagElementary( NodeAttribute< bool > &multible )
{
	if ( m_elementary )
	{
		if (m_injectiveRelation )
			multible[ m_subcentre ] = false;
		else
			multible[ m_subcentre ] = true;
		
		return; 
	}
	else
	{
		for( int i = 0; i <	m_numberOfSubbausteins; i++)
			 reinterpret_cast< RelationBaustein *>( m_subbausteins[ i ] )->SetMultibleFlagElementary( multible );
	}
}

	
//////////////////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////////////////


void RelationBaustein::SetRelation( NodeAttribute< int > &color, const Relation &relation )
{
	int i, entryPos, exitPos;
	Node first, second;

	ClearRelation(); // implemented by the derived class

	for ( i=0; i<relation.NumberOfPairs(); i++)
	{
		first = relation[ i ].First();

		for ( entryPos=0; entryPos< NumberOfEntries(); entryPos++)
			if ( m_entry[ entryPos ] == first )
				break;

		if ( entryPos >= NumberOfEntries() )
			return; //error - should never happend

		second = relation[ i ].Second();

		for ( exitPos=0; exitPos< NumberOfExits(); exitPos++)
			if ( m_exit[ exitPos ] == second )
				break;

		if ( exitPos >= NumberOfExits() )
			return; //error - should never happend

		ChangeRelation( entryPos, exitPos, color[ first ] ); // implemented by the derived class
	}
}

void RelationBaustein::SetMultibleFlags( NodeAttribute< bool > &multible, const Snet &snet )
{
	SetMultibleFlagElementary( multible ); // for subcentres only

	Node n, s = snet.FirstSubcentre();
	Edge e;

	while ( s )
	{
		bool multi = multible[ s ];

		// set multible flag for entries
		n = snet.FirstEntry( e, s );
		while ( n )
		{
			multible[ n ] = multi; 
			n = snet.NextEntry( e );
		}

		// set multible flag for exits
		n = snet.FirstExit( e, s );
		while ( n )
		{
			multible[ n ] = multi; 
			n = snet.NextExit( e );
		}

		s = snet.NextSubcentre( s );
	}
}


//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

void RelationBaustein::ComputePermutation( int *permutation )
{
	int i;  

	// initialise permutation
	for (i = 1; i<m_numberOfEntries; i++)
	{
		permutation[ i-1 ] = NO_EXIT;  // -1, because of dummy entry
	}

	// set mapping
	for (i = 1; i<m_numberOfExits; i++)
	{
		if ( m_exitEntry[ i ] != NO_ENTRY )  
		{
			permutation[ m_exitEntry[ i ]-1 ] = i - 1; // -1, because of dummy entry
		}
	}

}

void RelationBaustein::ComputeBenesMatrix( int size, int *permutation, Benes_t **benes )
{
	// compute Benes matrix

	int i, j;
	Benes_t *ben;

	*benes = ben = ben_new( size ); 

	for (i = 0; i<size; i++)
	{
		if ( permutation[ i ] != NO_EXIT )  
		{
				ben->IOrel[i] = permutation[ i ];
		}
	}

	ben_solve(ben); 

#ifdef TRACE_BENES
	printf("Benes%d\n", ben->relais->maxy * 2 );

	for (j=0;j<ben->relais->maxy;j++)
	{
		for(i=0;i<ben->relais->maxx;i++)
		{
			printf("%4X ",ben->relais->data[i][j]);
		}
		printf("\n");
	}

	printf("\n");
#endif //#ifdef TRACE_BENES

}


void RelationBaustein::DeleteBenesMatrix( Benes_t *benes )
{
	ben_delete( benes );
}
