////////////////////////////////////////////////////////////////////////////////
// File         : Snet.cpp
// Subsystem    : Graph
// Component    : S-net
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class Snet
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 10.09.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 07.01.04    Josef Huebl	 do not delete colored nodes in Snet::DeleteUnreachableNodes()
//                               otherwise start nodes of search pathes can be deleted
// 20.01.04    Josef Huebl    error correction in ChangeSubcentre()
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////


#define TRACE

#include "Snet.h"
#include "route.h"

#ifdef TRACE
#include "SnetPrint.h"
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Snet::Snet()
{
	m_sx = m_g.CreateNode(); // add subcenter for outer entries and exits
	m_nodeType[ m_sx ] = NODE_SUBCENTRE;
}


Snet::~Snet()
{

}

//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////


Snet& Snet::operator=( const Snet &snet)
{
	if (this != &snet)
		Copy( snet );

	return *this;
}


Node Snet::AddOuterEntry( void )
{
	Node n = m_g.CreateNode();
	m_g.CreateEdge( m_sx, n ); // Attention! An outer entry is an inner exit!
	m_nodeType[ n ] = NODE_EXIT;
	return n;
}


Node Snet::AddOuterExit( void )
{
	Node n = m_g.CreateNode();
	m_g.CreateEdge( n, m_sx ); // Attention! An outer exit is an inner entry!
	m_nodeType[ n ] = NODE_ENTRY;
	return n;
}


Node Snet::AddSubcentre( int numEntries, int numExits )
{
	int i;
	Node n, s = m_g.CreateNode();

	m_nodeType[ s ] = NODE_SUBCENTRE;

	for (i=0; i<numEntries; i++) 
	{
		n = m_g.CreateNode();
		m_g.CreateEdge( n, s );
		m_nodeType[ n ] = NODE_ENTRY;
	}

	for (i=0; i<numExits; i++) 
	{
		n = m_g.CreateNode();
		m_g.CreateEdge( s, n );
		m_nodeType[ n ] = NODE_EXIT;
	}

	return s;
}


void Snet::ChangeSubcentre( const Node &s, int numEntries, int numExits )
{
	int diff, i;
	Node n;
	Edge e;

	// entries
	diff = numEntries - NumberOfEntries( s );

	for (i=0; i < diff; i++) 
	{
		n = m_g.CreateNode();
		m_g.CreateEdge( n, s );
		m_nodeType[ n ] = NODE_ENTRY;
	}

	if ( diff < 0 )
	{
		// this entries remain
		e = s.FirstIngoingEdge();
		for (i=0; i < numEntries; i++) 
		{
			e = e.NextIngoingEdge();
		}

		// delete all other entries
		n = e.SourceNode();

		while ( e )
		{
			e = e.NextIngoingEdge();
			m_g.Delete( n );

		}
	}

	// if diff == 0; nothing to do

	// exits

	diff = numExits - NumberOfExits( s );

	for (i=0; i < diff; i++) 
	{
		n = m_g.CreateNode();
		m_g.CreateEdge( s, n );
		m_nodeType[ n ] = NODE_EXIT;
	}

	if ( diff < 0 )
	{
		// this exits remain
		e = s.FirstOutgoingEdge();
		for (i=0; i < numExits; i++) 
		{
			e = e.NextOutgoingEdge();
		}

		// delete all other exits
		n = e.TargetNode();

		while ( e )
		{
			e = e.NextOutgoingEdge();
			m_g.Delete( n );

		}
	}

	// if diff == 0; nothing to do
}


void Snet::DelOuterEntry( Node &y  )
{
	if ( m_nodeType[ y ] != NODE_EXIT) // Attention! An outer entries is an inner exits!
		return;

	m_g.Delete( y );
}


void Snet::DelOuterExit( Node &x  )
{
	if ( m_nodeType[ x ] != NODE_ENTRY) // Attention! An outer exit is an inner entry!
		return;

	m_g.Delete( x );
}



void Snet::DelSubcentre( Node &s  )
{
	Node n;

	if ( m_nodeType[ s ] != NODE_SUBCENTRE)
		return;
	
	// delete entries
	Edge edge = s.FirstIngoingEdge();

	while ( edge )  
	{
		n = edge.SourceNode() ;
		edge = edge.NextIngoingEdge();
		m_g.Delete( n );
	}

	// delete exits
	edge = s.FirstOutgoingEdge();

	while ( edge )  
	{
		n = edge.TargetNode() ;
		edge = edge.NextOutgoingEdge();
		m_g.Delete( n );
	}

	// delete subcentre
	m_g.Delete( s );
}


Node Snet::FirstSubcentre( void ) const
{
	Node n = m_g.FirstNode();

	while ( n )
	{
		if (( m_nodeType[ n ] == NODE_SUBCENTRE) && (n != m_sx))
			return n;
		else
			n = n.NextNode();
	}

	return n;
}


Node Snet::LastSubcentre( void ) const
{
	Node n = m_g.LastNode();

	while ( n )
	{
		if (( m_nodeType[ n ] == NODE_SUBCENTRE) && (n != m_sx))
			return n;
		else
			n = n.PreviousNode();
	}

	return n;
}


Node Snet::PrevSubcentre( const Node &nn) const 
{
	if (!nn)
		return nn;

	Node n = nn.PreviousNode();

	while ( n )
	{
		if (( m_nodeType[ n ] == NODE_SUBCENTRE) && (n != m_sx))
			return n;
		else
			n = n.PreviousNode();
	}

	return n;
}


Node Snet::NextSubcentre( const Node &nn) const 
{
	if (!nn)
		return nn;

	Node n = nn.NextNode();

	while ( n )
	{
		if (( m_nodeType[ n ] == NODE_SUBCENTRE) && (n != m_sx))
			return n;
		else
			n = n.NextNode();
	}

	return n;
}


Node Snet::SubcentreOf( const Node &n) const
{
	Edge e ;

	switch( GetNodeType( n ) )
	{
	case NODE_ENTRY:
		return n.FirstOutgoingEdge().TargetNode();
		break;

	case NODE_EXIT:
		return n.FirstIngoingEdge().SourceNode();
		break;

	default:
		return Node(); // invalid Node
		break;
	}
}


Node Snet::Replace( Node &s, const Snet &snet )
{
	Node sx, n, nOrig, LastNode, FirstNode;
	LastNode = m_g.CreateNode();  // mark end of list of nodes
	sx = m_g.AddCopyOf( snet.m_g );

	// copy attributes
	n = sx; 
	nOrig = snet.m_g.FirstNode(); 

	while ( n && nOrig )  
	{
		m_nodeType[ n ] = snet.m_nodeType[ nOrig ];
		n = n.NextNode();
		nOrig = nOrig.NextNode();
	}

	// build connections
	Edge e, e1, e2, e1s, e2s;
	Node sx_x, sx_y, s_x, s_y;

	s_y = FirstEntry( e1s, s );
	sx_x = FirstExit( e2, sx );

	while ( s_y && sx_x )
	{
		m_g.AddTransitiveHull( s_y, sx_x  );
		s_y = NextEntry( e1s );
		sx_x = NextExit( e2 );
	}

	sx_y = FirstEntry( e1, sx );
	s_x = FirstExit( e2s, s );

	while ( sx_y && s_x )
	{
		m_g.AddTransitiveHull( sx_y, s_x  );
		sx_y = NextEntry( e1 );
		s_x = NextExit( e2s );
	}

	DelSubcentre( s );
	DelSubcentre( sx );

	FirstNode = LastNode.NextNode();
	m_g.Delete( LastNode );
	return FirstNode; // newest created node
}




Edge Snet::Connect( const Node &x, const Node &y ) 
{
	NodeType tx, ty; 

	if ( !x || !y )
		return Edge(); // invalid edge; connection not possible

	tx = m_nodeType[ x ];
	ty = m_nodeType[ y ];

	if ( (tx != NODE_EXIT) || (ty != NODE_ENTRY) )
		return Edge(); // invalid edge; connection not allowed

	return m_g.CreateEdgeSingle( x, y );
}


void Snet::Disconnect( Edge &e ) 
{
	NodeType tx, ty; 

	tx = m_nodeType[ e.SourceNode() ];
	ty = m_nodeType[ e.TargetNode() ];

	if ( (tx != NODE_EXIT) || (ty != NODE_ENTRY) )
		return; // invalid edge; connection not allowed

	m_g.Delete( e );
	return;
}



Edge Snet::FirstConnection( void ) const 
{
	Edge e = m_g.FirstEdge();

	while ( e )
	{
		if (( m_nodeType[ e.SourceNode() ] == NODE_EXIT) && ( m_nodeType[ e.TargetNode() ] == NODE_ENTRY))
			return e;
		else
			e = e.NextEdge();
	}

	return e;
}


Edge Snet::NextConnection( const Edge &e) const 
{
	if (!e)
		return e;

	Edge ee = ee.NextEdge();

	while ( ee )
	{
		if (( m_nodeType[ ee.SourceNode() ] == NODE_EXIT) && ( m_nodeType[ ee.TargetNode() ] == NODE_ENTRY))
			return ee;
		else
			ee = ee.NextEdge();
	}

	return ee;
}


bool Snet::MakeTransitiveRec( void )
{
	bool changed = false;
	Node s = Root();
	Node x1, x2, y1, y2 ;
	Edge e, e1, e2, ee;

	while ( s )
	{
		x1 = FirstExit( e, s );

		while ( x1 )
		{
			e1 = x1.FirstOutgoingEdge();

				while ( e1 )
				{
					y1 = e1.TargetNode();
					e2 = x1.FirstOutgoingEdge();

					while ( e2 )
					{
						if ( e1 != e2 )
						{
							y2 = e2.TargetNode();
							ee = y1.FirstIngoingEdge();

							while ( ee )
							{
								x2 = ee.SourceNode();
								if ( !m_g.ExistingEdge( x2, y2 ) )
								{
									m_g.CreateEdge( x2, y2 );
									changed = true;
								}
								ee = ee.NextIngoingEdge();
							}
						}

						e2 = e2.NextOutgoingEdge();
					}

					e1 = e1.NextOutgoingEdge();
				}

			x1 = NextExit( e );
		}

		s = NextSubcentre( s );
	}

	return changed;
}


void Snet::TransComplement( Snet &complement, NodeAttribute< Node > &originNode ) // creates the transitive complement
{
	// Attention!
	// be very carefull when changing something.
	// The root node must always be the first created node
	// the ordering of in- and outgoing edges must not be destroyed
	// use the original nodes (after copy) for entries and exits whenever possible

	complement = *this;

	complement.MakeTransitive();

	// store reference to original Node
	Node nn, n = complement.m_g.FirstNode();
	Node nOrig = m_g.FirstNode();

	while (n)
	{
		originNode[ n ] = nOrig;
		n = n.NextNode();
		nOrig = nOrig.NextNode();
	}

	//compute outer entries; but do not destroy the ordering of outer entries and exits
	Edge e1, e2, e3, enext;
	Node nNew, nPrev, nNext, next, root = complement.Root();

	complement.m_g.AddStarBetween( root );  // add entries and exits for root

	// check all outer entries

	Edge e = root.FirstOutgoingEdge();

	while ( e )
	{
		nPrev = e.TargetNode(); 
		n = nPrev.FirstOutgoingEdge().TargetNode(); // outer entry 

		if ( complement.m_g.NumberOfOutgoingEdges( n ) != 0 ) // check outer entry
		{
			nNext = n.FirstOutgoingEdge().TargetNode(); 
			next = n;
			
			int numberOfSubcentres = 0; 

			while (( complement.m_g.NumberOfIngoingEdges( next ) == 1 ) && ( complement.m_g.NumberOfOutgoingEdges( next ) == 1 ) && (next != root) )
			{
				next = next.FirstOutgoingEdge().TargetNode();
				if ( complement.m_nodeType[ next ] == NODE_SUBCENTRE )
					numberOfSubcentres++;
			}
			if ( (( next == root) && ( numberOfSubcentres % 2 == 1)) ||
				(( next != root) && ( complement.m_nodeType[ next ] == NODE_SUBCENTRE ) && (complement.m_g.NumberOfOutgoingEdges( next ) == 0)) ||
				(( complement.m_nodeType[ next ] != NODE_SUBCENTRE ) && (complement.m_g.NumberOfOutgoingEdges( next ) != 0)) )
			{
				// add new outer entry
				complement.m_g.Delete( nPrev.FirstOutgoingEdge() );
				nNew = complement.m_g.CreateNode();
				complement.m_g.CreateEdge( nPrev, nNew );
				complement.m_g.CreateEdge( nNew, n );
			}
			else
			{
				// delete outer entry
				complement.m_g.CreateEdge( nPrev, nNext );
				complement.m_g.Delete( n );
			}
		}
		e = e.NextOutgoingEdge();
	}


	// check all outer exits
	e = root.FirstIngoingEdge();

	while ( e )
	{
		nNext = e.SourceNode(); 
		n = nNext.FirstIngoingEdge().SourceNode(); // outer exit

		if ( complement.m_g.NumberOfIngoingEdges( n ) != 0 ) // check outer exit
		{
			nPrev = n.FirstIngoingEdge().SourceNode(); 
			next = n;
			
			int numberOfSubcentres = 0;  
			while (( complement.m_g.NumberOfIngoingEdges( next ) == 1 ) && ( complement.m_g.NumberOfIngoingEdges( next ) == 1 ) && (next != root) )
			{
				next = next.FirstIngoingEdge().SourceNode();
				if ( complement.m_nodeType[ next ] == NODE_SUBCENTRE )
					numberOfSubcentres++;
			}

			if ( (( next == root) && (numberOfSubcentres % 2 == 1)) ||
				(( next != root) && ( complement.m_nodeType[ next ] == NODE_SUBCENTRE ) && (complement.m_g.NumberOfIngoingEdges( next ) == 0)) ||
				(( complement.m_nodeType[ next ] != NODE_SUBCENTRE ) && (complement.m_g.NumberOfIngoingEdges( next ) != 0)) )
			{
				// add new outer exit
				complement.m_g.Delete( nNext.FirstIngoingEdge() );
				nNew = complement.m_g.CreateNode();
				complement.m_g.CreateEdge( n, nNew );
				complement.m_g.CreateEdge( nNew, nNext );
			}
			else
			{
				// delete outer exit
				complement.m_g.CreateEdge( nPrev, nNext );
				complement.m_g.Delete( n );
			}
		}
		e = e.NextIngoingEdge();
	}

	complement.m_g.DeleteStarBetween( root );

	// set node type for all entries and exits
	complement.SetNodeTypes( root );

	// mark all old subcentres as replacable
	NodeAttribute< bool > isReplaceable( false );

	Node ss, s = complement.FirstSubcentre();
	while (s )
	{
		isReplaceable[ s ] = true;
		s = complement.NextSubcentre( s );
	}

	// replace inbetween edges with subcentres
	e = complement.m_g.FirstEdge();

	while ( e )
	{
		if (  (complement.m_nodeType[ e.SourceNode() ] == NODE_EXIT) && (complement.m_nodeType[ e.TargetNode() ] == NODE_ENTRY) )
		{
			Edge ee = e.PreviousEdge(); // store because e becomes deleted
			s = complement.m_g.AddStar( e.TargetNode(), e.SourceNode() );
			complement.m_nodeType[ s ] = NODE_SUBCENTRE;
			complement.m_g.DeleteTransitiveHull( s );
			if ( ee )
				e = ee.NextEdge();
			else
				e = complement.m_g.FirstEdge(); //  e was the first edge
		}
		else
			e = e.NextEdge();
	}

	// replace subcentres
	s = complement.FirstSubcentre();
	while (s )
	{
		if ( isReplaceable[ s ] )
		{
			complement.m_g.AddTransitiveHull( s );
			ss = complement.NextSubcentre( s );  // s becomed deleted
			complement.m_g.Delete( s );
		}

		if ( s )
			s = complement.NextSubcentre( s );
		else
			s = ss; // s has been deleted
	}

	// search for free entries and exits
	n = complement.m_g.FirstNode();
	while ( n ) 
	{
		if ( ( complement.m_nodeType[ n ] == NODE_ENTRY) && ( complement.m_g.NumberOfIngoingEdges( n ) == 0 ) && (n.FirstOutgoingEdge().TargetNode() != root))
		{
			s = complement.m_g.CreateNode();
			complement.m_nodeType[ s ] = NODE_SUBCENTRE;
			complement.m_g.CreateEdge( s, n );
		}
		else if ( ( complement.m_nodeType[ n ] == NODE_EXIT) && ( complement.m_g.NumberOfOutgoingEdges( n ) == 0 ) && (n.FirstIngoingEdge().SourceNode() != root) )
		{
			s = complement.m_g.CreateNode();
			complement.m_nodeType[ s ] = NODE_SUBCENTRE;
			complement.m_g.CreateEdge( n, s );
		}

		n = n.NextNode();
	}


	// set node types for entries and exits	
	s = complement.FirstSubcentre();
	while (s )
	{
		complement.SetNodeTypes( s );
		s = complement.NextSubcentre( s );
	}

}


bool Snet::ComputeRelations( NodeAttribute< int > &color, NodeAttribute< bool > &multible, NodeAttribute< Relation > &relation )
{
	int i, defaultColor, invalidColor;
	Node n, nOrig;
	Edge e, eOrig;
#ifdef TRACE
	SnetPrint complement( "Complement" );
#else
	Snet complement;
#endif

	NodeAttribute< Node > originNode;

	// compute the transitive complement
	TransComplement( complement, originNode ); 

#ifdef TRACE
	complement.Renumber();
	complement.Print();
#endif

	NodeAttribute< int > componentNumber( 0 );
	int numberOfComponents = complement.MarkComponents( componentNumber );

	int entriesPerComponet[ 100 ];

	for ( i=0; i<100; i++)
		entriesPerComponet[ i ] = 0;

	n = complement.FirstOuterEntry( e );

	while ( n )
	{
		entriesPerComponet[ componentNumber[ n ] ]++;
		n = complement.NextOuterEntry( e );
	}

	int max = 0;

	for ( i=0; i<100; i++)
		if (entriesPerComponet[ i ] > max )
			max = entriesPerComponet[ i ];



	// copy or set multible flag for all entries and exits
	// for all other nodes it is set to true by default
	NodeAttribute< bool > complementMultible( true ); 

	Node s;

	s = complement.FirstSubcentre();

	while ( s )
	{
		// copy multible flag for entries
		n = complement.FirstEntry( e, s );
		while ( n )
		{
			bool multi = multible[ originNode[ n ] ]; 

			if (multi != true) // copy value only if it is different from default value
				complementMultible[ n ] = true; 

			n = complement.NextEntry( e );
		}

		// copy multible flag for exits
		n = complement.FirstExit( e, s );
		while ( n )
		{
			bool multi = multible[ originNode[ n ] ]; 

			if (multi != true) // copy value only if it is different from default value
				complementMultible[ n ] = true; 

			n = complement.NextExit( e );
		}

		s = complement.NextSubcentre( s );
	}

	// compute invalid color (unused color )
	defaultColor = color.GetDefaultValue();
	invalidColor = defaultColor;
	nOrig = FirstOuterExit( eOrig );
	while ( nOrig )
	{
		if ( color[ nOrig ] >= invalidColor )
		{
			invalidColor = color[ nOrig ] + 1;
		}
		nOrig = NextOuterExit( eOrig );
	}

	// prepair SourceNode and targets for routing
	NodeAttribute< int > complementColor( defaultColor ); 
	NodeAttribute< int > complementID; //for debugging only
	NodeVector sources;
	NodeVector targets;

	n = complement.FirstOuterEntry( e );
	nOrig = FirstOuterEntry( eOrig );
	while ( n )
	{
		int col = color[ nOrig ]; // for debugging only

		if ( color[ nOrig ] != defaultColor )
		{
			sources.AddLast( n );
			complementColor[ n ] = color[ nOrig ];
		}
		else
		{
			complementColor[ n ] = invalidColor; // otherwise the rooter doesn't stop before Root()
			// But this is not necessary if complement.DeleteUnreachableNodes() is called for optimisation later
		}

		n = complement.NextOuterEntry( e );
		nOrig = NextOuterEntry( eOrig );
	}

	n = complement.FirstOuterExit( e );
	nOrig = FirstOuterExit( eOrig );
	while ( n )
	{
		if ( color[ nOrig ] != defaultColor )
		{
			targets.AddLast( n );
			complementColor[ n ] = color[ nOrig ];
		}
		n = complement.NextOuterExit( e );
		nOrig = NextOuterExit( eOrig );
	}

	// you can set in comment the next call of DeleteUnreachableNodes(),
	// if you are not shure, whether it works correct.
	// Note: The call of DeleteUnreachableNodes() destroys the S-Net partly. 
	complement.DeleteUnreachableNodes( complementColor, defaultColor, invalidColor );

	// prepair pathes for routing
	NodePathVector pathes( targets.NumberOfNodes() );

	for ( i=0; i<targets.NumberOfNodes(); i++)
		pathes[ i ].AddLast( targets[ i ] );

	// search colored trees
	Router router( complement.m_g, complementColor, complementMultible, componentNumber, complementID );
	bool ret = router.SearchColoredPathes( pathes );

#ifdef TRACE
	int np = pathes.NumberOfPathes(); // for debugging only

	for ( i=0; i<pathes.NumberOfPathes(); i++)
	{
		printf( "path[ %d ]", i );
		NodePathPrint pp( complement, pathes[ i ], "founded Path in Complement" );
		pp.Print();
	}
#endif

	if (ret )
		ret = router.CheckPathes( sources,targets, pathes ); // for debugging only

	// map nodes to original S-net and sort relations
	for ( i=0; i<pathes.NumberOfPathes(); i++)
	{
		Node s, nNext;
		NodePath *path = &(pathes[ i ]);

		int nn = path->NumberOfNodes(); // for debugging only

		n = path->First();
		nNext = path->Next();

		while( nNext )
		{
			if (( complement.GetNodeType( n ) == NODE_ENTRY ) && ( complement.GetNodeType( nNext ) == NODE_EXIT ) 
				&& originNode[ nNext ] && originNode[ n ] ) // because of new outer entries and exits
			{
				color[ originNode[ n ] ] = complementColor[ n ];
				color[ originNode[ nNext ] ] = complementColor[ nNext ];
				NodePair nodePair( originNode[ nNext ], originNode[ n ] );
				s = SubcentreOf( originNode[ n ] );
				relation[ s ].AddLast( nodePair );
			}
			n = nNext;
			nNext = path->Next();
		}
	}

	return ret;
}

	
void Snet::DeleteUnreachableNodes( NodeAttribute< int > &color, int defaultColor, int invalidColor )
{
    Node entry, exit, next;
	Edge e;

	// delete unused outer entries
	entry = FirstOuterEntry( e );

	while ( entry )
	{
		if (( color[ entry ] == defaultColor ) || ( color[ entry ] == invalidColor ) )
		{
			next = NextOuterEntry( e );
			m_g.Delete( entry );
			entry = next;
		}
		else
			entry = NextOuterEntry( e );
	}

	// delete unused outer exit
	exit = FirstOuterExit( e );

	while ( exit )
	{
		if ( color[ exit ] == defaultColor )
		{
			next = NextOuterExit( e );
			m_g.Delete( exit );
			exit = next;
		}
		else
			exit = NextOuterExit( e );
	}


	// delete nodes which have no ancestor or no predecessor is not root adn uncolored

	Node n;
	Edge eIn, eOut;
	unsigned int deleted = 0; // for debugging only

	bool found = true;

	while ( found )
	{
		found = false;
		n = m_g.FirstNode();

		while ( n )
		{
			eIn = n.FirstIngoingEdge();
			eOut = n.FirstOutgoingEdge();

			if (( !eIn || !eOut) && (n != Root()) && (( color[ n ] == defaultColor ) || ( color[ n ] == invalidColor ) ))
			{
				next = n.NextNode();
				m_g.Delete( n );
				n = next;
				found = true;
				deleted++;
			}
			else
				n = n.NextNode();
		}
	}
}



//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

void Snet::Copy( const Snet &snet)
{
	Node n, nOrig;

	m_g = snet.m_g;

	// copy attributes
	n = m_g.FirstNode(); 
	nOrig = snet.m_g.FirstNode(); 

	while ( n && nOrig )  
	{
		m_nodeType[ n ] = snet.m_nodeType[ nOrig ];

		if ( nOrig == snet.m_sx )
			m_sx = n;

		n = n.NextNode();
		nOrig = nOrig.NextNode();
	}
}


int Snet::MarkComponents( NodeAttribute< int > &componentNumber )
{
	int numberOfcomponents = 0;

	componentNumber[ Root() ] = -1;

	Node n = m_g.FirstNode();

	while ( n )
	{
		if (componentNumber[ n ] == 0)
		{
			MarkNeighbours( n, componentNumber, ++numberOfcomponents ); 
		}
		n = n.NextNode();
	}
	return numberOfcomponents;
}


//////////////////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////////////////

void Snet::SetNodeTypes( const Node &s ) // set m_nodeType for a subcentre and its entries and exits
{
	m_nodeType[ s ] = NODE_SUBCENTRE;

	Edge e = s.FirstIngoingEdge();
	while ( e )
	{
		m_nodeType[ e.SourceNode() ] = NODE_ENTRY;
		e = e.NextIngoingEdge();
	}

	e = s.FirstOutgoingEdge();
	while ( e )
	{
		m_nodeType[ e.TargetNode() ] = NODE_EXIT;
		e = e.NextOutgoingEdge();
	}

}


void Snet::MarkNeighbours( Node n, NodeAttribute< int > &componentNumber, int componentNr )
{
	Edge e;

	if (componentNumber[ n ] == 0)
	{
		componentNumber[ n ] = componentNr;

		e = n.FirstIngoingEdge();

		while ( e )
		{
			MarkNeighbours( e.SourceNode(), componentNumber, componentNr );
			e = e.NextIngoingEdge();
		}

		e = n.FirstOutgoingEdge();

		while ( e )
		{
			MarkNeighbours( e.TargetNode(), componentNumber, componentNr );
			e = e.NextOutgoingEdge();
		}
	}
}

