////////////////////////////////////////////////////////////////////////////////
// File         : indexattr.h
// Subsystem    : Graph
// Component    : index based attributation
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : definition of index based attributation
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 05.08.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
// 
// Some classes are implemented, that allow to manage a index based attributation.
// Elements that use indexed attributs must be derived from class AttributedElement.
// An attribute is defined by the template class Attribute< ElementType, AttributeType >.
//
// Example: 
// class Node is derived from AttributedElement.
// Attribute< Node, bool > visible declares a boolean node attribute named visible.
// If n is of type Node, visible[ n ] is the attribute value.
//
// For a detailed describtion see the document Attributverwaltung.doc. 
//     
////////////////////////////////////////////////////////////////////////////////


#ifndef INDEXATTR_H
#define INDEXATTR_H

#ifdef WIN32
#pragma warning( disable : 4786 )  // Disable warning message C4786
#endif

#include <vector>
#include <list>

typedef unsigned long Index;
typedef Index AttributeID;
typedef Index AttributeValueIndex;

const Index					NONE_VALID_INDEX = -1;
const AttributeID			NONE_VALID_ATTRIBUTE_ID = NONE_VALID_INDEX;
const AttributeValueIndex	NONE_VALID_ATTRIBUTE_VALUE_INDEX = NONE_VALID_INDEX;

// ****************************** IndexAdmin ************************************

class IndexAdmin
{
public:
	IndexAdmin() : m_nextUnusedIndex(0), m_reusableIndexes(0) {};
	~IndexAdmin() { m_reusableIndexes.clear(); };
	Index NewIndex( void ); 
	Index MaxIndex( void ); 
	void DelIndex(Index index); 
	void ReInitialize(Index nextIndexInUse ); 
	void ClearIndexes( void ); 
	
private:
	Index m_nextUnusedIndex;
	std::vector<Index> m_reusableIndexes;			// indexes for recycling
};

// ****************************** AttributeIndexAdmin ************************************

class AttributeIndexAdmin
{
public:
	AttributeIndexAdmin() : m_attributeValueIndexAdmin(0) {};
	~AttributeIndexAdmin() { m_attributeValueIndexAdmin.clear(); };
	AttributeID NewAttributeID( void );			
	AttributeID CopyAttributeID( AttributeID originalID );			
	AttributeID MaxAttributeID( void );			
	void DelAttributeID( AttributeID id );			
	AttributeValueIndex NewAttributeValueIndex( AttributeID id ); 
	AttributeValueIndex MaxAttributeValueIndex( AttributeID id ); 
	void DelAttributeValueIndex( AttributeID id, AttributeValueIndex valueIndex ); 
	void ReInitializeAttributeValueIndexes( AttributeID id, AttributeValueIndex nextValueIndexInUse ); 
	void ClearAttributeValueIndexes( AttributeID id ); 
	void Clear( void ); 
private:
	IndexAdmin m_attributeIDadmin;
	std::vector<IndexAdmin> m_attributeValueIndexAdmin;
};


struct IndexedAttribute 
{
	AttributeValueIndex attributeValueIndex;		// zur Speicherplatz-Optimierung
	bool virgin;	// == true, if never gots a real attribute value
};

class AttributeVectorTable;

// ****************************** AttributeVector ************************************

class AttributeVector
{
public:
	AttributeVector( bool *validPointerPtr ) : m_attributeVectorTablePtr( NULL ), m_validPointerPtr( validPointerPtr )  {};
	IndexedAttribute &GetIndexedAttribute( AttributeID id ) { return m_indexedAttributes[ id ]; };
	void NewIndexedAttribute( AttributeID id ); 
	void CopyIndexedAttribute( AttributeID idOriginal, AttributeID idCopy ); 
	void DelIndexedAttribute( AttributeID id ); 
	void ResetIndexedAttribute( AttributeID id ) {m_indexedAttributes[ id ].virgin = true;};
	void UseIndexedAttribute( AttributeID id ) {m_indexedAttributes[ id ].virgin = false;};
	void AddAllAttributes( AttributeVectorTable *table ); 
	void DelAllAttributes( void ); 
	AttributeVectorTable *GetAttributeVectorTablePtr( void ) { return m_attributeVectorTablePtr; };

private:
	std::vector<IndexedAttribute> m_indexedAttributes;
	AttributeVectorTable *m_attributeVectorTablePtr;
	bool *m_validPointerPtr; // for destructor
};

// ****************************** AttributeVectorTable ************************************

class AttributeVectorTable
{
public:
	virtual ~AttributeVectorTable();
	std::list<AttributeVector>::iterator NewAttributeVector( bool *validPointerPtr );
	void DelAttributeVector( std::list<AttributeVector>::iterator );
	void DelAllAttributeVectors( void );
	AttributeID NewAttribute( AttributeVectorTable **attributeVectorTablePtrPtr );
	AttributeID CopyAttribute( AttributeID originalID, AttributeVectorTable **copyAttributeVectorTablePtrPtr );
	void DelAttribute( AttributeID id );
	void ReinitialiseAttribute( AttributeID id );
	AttributeValueIndex MaxAttributeValueIndex( AttributeID id ) { return m_attributeIndexAdmin.MaxAttributeValueIndex( id ); };
	AttributeIndexAdmin &GetAttributeIndexAdmin(void) { return m_attributeIndexAdmin; }; 

	Index GetNumberOfElements( void ) { return m_attributeVectors.size(); };
	bool SetToFirstAttributeVector( void ) { m_it = m_attributeVectors.begin(); if ( m_it == m_attributeVectors.end()) return false; else return true; };
	bool SetToNextAttributeVector( void ) { if ( ++m_it == m_attributeVectors.end()) return false; else return true; };
	IndexedAttribute GetCurrentIndexedAttribute( AttributeID id ) { return (m_it->GetIndexedAttribute( id )); };
	void TouchCurrentIndexedAttribute( AttributeID id ) { m_it->GetIndexedAttribute( id ).virgin = false; };

private:
	std::vector<AttributeVectorTable **> m_attributeVectorTablePtrPtr; // for independent destruction
	AttributeIndexAdmin m_attributeIndexAdmin;
	std::list<AttributeVector> m_attributeVectors;
	std::list<AttributeVector>::iterator m_it;
};


// ****************************** AttributedElement ************************************
// Elements that use indexed attributs must be derived from AttributedElement

class AttributedElement
{
public:
	AttributedElement( ) : m_validPointer(false) {};
	void NewAttributeVector( AttributeVectorTable *ptr );
	void DelAttributeVector( void );
	IndexedAttribute &GetIndexedAttribute( AttributeID id ) const { return m_attributeVectorPtr->GetIndexedAttribute( id ); };
	std::list<AttributeVector>::iterator GetAttributeVectorPtr( void ) const { return m_attributeVectorPtr; };
	void SetAttributeVectorPtr( std::list<AttributeVector>::iterator ptr ) { m_validPointer = true; m_attributeVectorPtr = ptr; };

private:
	mutable bool m_validPointer;
	mutable std::list<AttributeVector>::iterator m_attributeVectorPtr;
};

// ****************************** AttributeBase ************************************
// Attibute classes must be derived from AttributeBase

class AttributeBase
{
protected:
	AttributeBase() : m_attributID( NONE_VALID_ATTRIBUTE_ID ), m_attributeVectorTablePtr( NULL ) {}; 
	virtual ~AttributeBase() { if (m_attributeVectorTablePtr != NULL) m_attributeVectorTablePtr->DelAttribute( m_attributID ); }; 
	AttributeValueIndex CheckAttributeValue( const std::list<AttributeVector>::iterator vectorPtr ) const;
	IndexedAttribute GetIndexedAttribute( const std::list<AttributeVector>::iterator vectorPtr ) const;
	virtual void ResetToDefaultValue( AttributeValueIndex index ) const = 0;  // pure virtual; must be implemented by the attribute class
	virtual void ResizeUpTo( AttributeValueIndex index ) const = 0;  // pure virtual; must be implemented by the attribute class

public:
	void ReinitialiseValues( void ){ if (m_attributeVectorTablePtr != NULL) m_attributeVectorTablePtr->ReinitialiseAttribute( m_attributID ); }; 

protected:
	mutable AttributeID m_attributID;
	mutable AttributeVectorTable *m_attributeVectorTablePtr; // for destructor
};

/// *********************** Template class Attribute **********************************

template< typename ElementType, typename AttributeType >
class Attribute : public AttributeBase
{
public:
	Attribute(): m_attributeDefaultValue( AttributeType() ){};
	Attribute( AttributeType defaultValue ) : m_attributeDefaultValue(defaultValue) {};
	Attribute( const Attribute &a ) { CopyValues( a ); };
	Attribute& operator=(const Attribute &a) { CopyValues( a ); return *this; };
	friend bool operator==(const Attribute &a1, const Attribute &a2) {return a1.Equal( a2 ); };
	friend bool operator!=(const Attribute &a1, const Attribute &a2) { return !(a1 == a2); };
	void SetDefaultValue( ElementType value );
	AttributeType GetDefaultValue( void ) const { return m_attributeDefaultValue; };
	AttributeType& operator [](const ElementType element);
	const AttributeType& operator [](const ElementType element) const;

private:
	AttributeType GetValue( const IndexedAttribute &ia) const { if (ia.virgin) return m_attributeDefaultValue; else return m_attributeValues[ ia.attributeValueIndex ]; };	
	void ResetToDefaultValue( AttributeValueIndex index ) const; 
	void ResizeUpTo( AttributeValueIndex index ) const; 
	void CopyValues(const Attribute &attribute); // deletes old attribute from table, creates a new attribute id and copies all values
	bool Equal( const Attribute &a) const;

	AttributeType		m_attributeDefaultValue;  
	mutable std::vector<AttributeType>	m_attributeValues;	
};


// *********************** Template class Attribute methods **********************************


template< typename ElementType, typename AttributeType >
void Attribute< ElementType, AttributeType >::ResizeUpTo( AttributeValueIndex index ) const
{
	if (m_attributeValues.size() <= index) 
		m_attributeValues.resize( index + 1, m_attributeDefaultValue ); 
}


template< typename ElementType, typename AttributeType >
void Attribute< ElementType, AttributeType >::ResetToDefaultValue( AttributeValueIndex index ) const
{
	if (m_attributeValues.size() <= index) 
		m_attributeValues.resize( index + 1, m_attributeDefaultValue );
	else
		m_attributeValues[ index ] = m_attributeDefaultValue;
}


template< typename ElementType, typename AttributeType >
void Attribute< ElementType, AttributeType >::SetDefaultValue( ElementType value )
{
	if (m_attributeDefaultValue == value)
		return;

	if (m_attributeVectorTablePtr == NULL)
	{
		m_attributeDefaultValue = value;
		return;
	}

	Index numberOfElements = m_attributeVectorTablePtr->GetNumberOfElements();

	if (numberOfElements == 0)
	{
		m_attributeDefaultValue = value;
		return;
	}

	IndexedAttribute ia;
	bool next = m_attributeVectorTablePtr->SetToFirstAttributeVector();

	while ( next )
	{
		ia = m_attributeVectorTablePtr->GetIndexAttribute( m_attributID );

		if ( ia.virgin )
		{
			ResetToDefaultValue( ia.attributeValueIndex );
			TouchCurrentIndexedAttribute( m_attributID );
		}

		next = m_attributeVectorTablePtr->SetToNextAttributeVector();
	}

	m_attributeDefaultValue = value;
}


template< typename ElementType, typename AttributeType >
AttributeType &Attribute< ElementType, AttributeType >::operator [](const ElementType element)
{
	AttributeValueIndex valueIndex = CheckAttributeValue( element.GetAttributeVectorPtr() );
	return m_attributeValues[ valueIndex ];
}


template< typename ElementType, typename AttributeType >
const AttributeType &Attribute< ElementType, AttributeType >::operator [](const ElementType element) const
{
	AttributeValueIndex valueIndex = CheckAttributeValue( element.GetAttributeVectorPtr() );
	return m_attributeValues[ valueIndex ];
}


template< typename ElementType, typename AttributeType >
void Attribute< ElementType, AttributeType >::CopyValues(const Attribute &attribute)
{
	if (m_attributeVectorTablePtr != NULL )
		m_attributeVectorTablePtr->DelAttribute( m_attributID );  // attribute is already in use

	m_attributeDefaultValue = attribute.m_attributeDefaultValue;  
	m_attributeValues = attribute.m_attributeValues;  
	m_attributeVectorTablePtr = attribute.m_attributeVectorTablePtr;  

	if ( m_attributeVectorTablePtr != NULL )
		m_attributID = m_attributeVectorTablePtr->CopyAttribute( attribute.m_attributID, &m_attributeVectorTablePtr );
	else
		m_attributID = attribute.m_attributID;  
}


template< typename ElementType, typename AttributeType >
bool Attribute< ElementType, AttributeType >::Equal( const Attribute &a) const
{
	// m_attributeDefaultValue and m_attributID are allowed to be different

	if (m_attributeVectorTablePtr != a.m_attributeVectorTablePtr)
		return false;

	if ((m_attributeVectorTablePtr == NULL) && (a.m_attributeVectorTablePtr == NULL))
		return true;

	Index numberOfElements = m_attributeVectorTablePtr->GetNumberOfElements();

	if (numberOfElements == 0)
		return true;        // no elements

	IndexedAttribute ia1, ia2;
	bool next = m_attributeVectorTablePtr->SetToFirstAttributeVector();

	while ( next )
	{
		ia1 = m_attributeVectorTablePtr->GetCurrentIndexedAttribute( m_attributID );
		ia2 = m_attributeVectorTablePtr->GetCurrentIndexedAttribute( a.m_attributID );

		if ( GetValue( ia1 ) != a.GetValue(ia2) )
			return false;

		next = m_attributeVectorTablePtr->SetToNextAttributeVector();
	}

	return true; // all are equal
}


#ifndef WIN32
// GNU handles bool different

template< typename ElementType >
class Attribute< ElementType, bool  > : public Attribute< ElementType, char >
{
public:
	Attribute() : Attribute< ElementType, char >() {};
	Attribute( bool defaultValue ) : Attribute< ElementType, char >( defaultValue ) {};
};

#endif

#endif // #ifdef INDEXATTR_H

